@@ -24,12 +24,8 @@
 //! We will use the following choices in this example:
 //!
 //! ```
-//! use voprf::CipherSuite;
-//! struct Default;
-//! impl CipherSuite for Default {
-//!     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//!     type Hash = sha2::Sha512;
-//! }
+//! type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! type Hash = sha2::Sha512;
 //! ```
 //!
 //! ## Modes of Operation
@@ -56,17 +52,13 @@
 //! client evaluations.
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! use voprf::NonVerifiableServer;
 //! use rand::{rngs::OsRng, RngCore};
 //!
 //! let mut server_rng = OsRng;
-//! let server = NonVerifiableServer::<Default>::new(&mut server_rng)
+//! let server = NonVerifiableServer::<Group, Hash>::new(&mut server_rng)
 //!    .expect("Unable to construct server");
 //! ```
 //!
@@ -79,18 +71,14 @@
 //! step of the VOPRF protocol.
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! use voprf::NonVerifiableClient;
 //! use rand::{rngs::OsRng, RngCore};
 //!
 //! let mut client_rng = OsRng;
-//! let client_blind_result = NonVerifiableClient::<Default>::blind(
-//!     b"input",
+//! let client_blind_result = NonVerifiableClient::<Group, Hash>::blind(
+//!     b"input".to_vec(),
 //!     &mut client_rng,
 //! ).expect("Unable to construct client");
 //! ```
@@ -104,28 +92,23 @@
 //! [EvaluationElement] to be sent to the client.
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! # use voprf::NonVerifiableClient;
 //! # use rand::{rngs::OsRng, RngCore};
 //! #
 //! # let mut client_rng = OsRng;
-//! # let client_blind_result = NonVerifiableClient::<Default>::blind(
-//! #     b"input",
+//! # let client_blind_result = NonVerifiableClient::<Group, Hash>::blind(
+//! #     b"input".to_vec(),
 //! #     &mut client_rng,
 //! # ).expect("Unable to construct client");
 //! # use voprf::NonVerifiableServer;
 //! # let mut server_rng = OsRng;
-//! # let server = NonVerifiableServer::<Default>::new(&mut server_rng)
+//! # let server = NonVerifiableServer::<Group, Hash>::new(&mut server_rng)
 //! #   .expect("Unable to construct server");
-//! use voprf::Metadata;
 //! let server_evaluate_result = server.evaluate(
 //!     client_blind_result.message,
-//!     &Metadata::none(),
+//!     None,
 //! ).expect("Unable to perform server evaluate");
 //! ```
 //!
@@ -133,40 +116,33 @@
 //!
 //! In the final step, the client takes as input the message from
 //! [NonVerifiableServer::evaluate] (an [EvaluationElement]), and runs
-//! [NonVerifiableClient::finalize] to produce a
-//! [NonVerifiableClientFinalizeResult], which consists of an
-//! output for the protocol.
+//! [NonVerifiableClient::finalize] to produce an output for the protocol.
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! # use voprf::NonVerifiableClient;
 //! # use rand::{rngs::OsRng, RngCore};
 //! #
 //! # let mut client_rng = OsRng;
-//! # let client_blind_result = NonVerifiableClient::<Default>::blind(
-//! #     b"input",
+//! # let client_blind_result = NonVerifiableClient::<Group, Hash>::blind(
+//! #     b"input".to_vec(),
 //! #     &mut client_rng,
 //! # ).expect("Unable to construct client");
 //! # use voprf::NonVerifiableServer;
 //! # let mut server_rng = OsRng;
-//! # let server = NonVerifiableServer::<Default>::new(&mut server_rng)
+//! # let server = NonVerifiableServer::<Group, Hash>::new(&mut server_rng)
 //! #   .expect("Unable to construct server");
 //! # let server_evaluate_result = server.evaluate(
 //! #     client_blind_result.message,
-//! #     &Metadata::none(),
+//! #     None,
 //! # ).expect("Unable to perform server evaluate");
-//! use voprf::Metadata;
 //! let client_finalize_result = client_blind_result.state.finalize(
 //!     server_evaluate_result.message,
-//!     &Metadata::none(),
+//!     None,
 //! ).expect("Unable to perform client finalization");
 //!
-//! println!("VOPRF output: {:?}", client_finalize_result.output.to_vec());
+//! println!("VOPRF output: {:?}", client_finalize_result.to_vec());
 //! ```
 //!
 //! ## Verifiable Mode
@@ -189,17 +165,13 @@
 //! client evaluations.
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! use voprf::VerifiableServer;
 //! use rand::{rngs::OsRng, RngCore};
 //!
 //! let mut server_rng = OsRng;
-//! let server = VerifiableServer::<Default>::new(&mut server_rng)
+//! let server = VerifiableServer::<Group, Hash>::new(&mut server_rng)
 //!    .expect("Unable to construct server");
 //!
 //! // To be sent to the client
@@ -219,18 +191,14 @@
 //! step of the VOPRF protocol.
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! use voprf::VerifiableClient;
 //! use rand::{rngs::OsRng, RngCore};
 //!
 //! let mut client_rng = OsRng;
-//! let client_blind_result = VerifiableClient::<Default>::blind(
-//!     b"input",
+//! let client_blind_result = VerifiableClient::<Group, Hash>::blind(
+//!     b"input".to_vec(),
 //!     &mut client_rng,
 //! ).expect("Unable to construct client");
 //! ```
@@ -244,29 +212,24 @@
 //! [EvaluationElement] to be sent to the client along with a proof.
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! # use voprf::VerifiableClient;
 //! # use rand::{rngs::OsRng, RngCore};
 //! #
 //! # let mut client_rng = OsRng;
-//! # let client_blind_result = VerifiableClient::<Default>::blind(
-//! #     b"input",
+//! # let client_blind_result = VerifiableClient::<Group, Hash>::blind(
+//! #     b"input".to_vec(),
 //! #     &mut client_rng,
 //! # ).expect("Unable to construct client");
 //! # use voprf::VerifiableServer;
 //! # let mut server_rng = OsRng;
-//! # let server = VerifiableServer::<Default>::new(&mut server_rng)
+//! # let server = VerifiableServer::<Group, Hash>::new(&mut server_rng)
 //! #   .expect("Unable to construct server");
-//! use voprf::Metadata;
 //! let server_evaluate_result = server.evaluate(
 //!     &mut server_rng,
 //!     client_blind_result.message,
-//!     &Metadata::none(),
+//!     None,
 //! ).expect("Unable to perform server evaluate");
 //! ```
 //!
@@ -275,43 +238,36 @@
 //! In the final step, the client takes as input the message from
 //! [VerifiableServer::evaluate] (an [EvaluationElement]),
 //! the proof, and the server's public key, and runs
-//! [VerifiableClient::finalize] to produce a
-//! [VerifiableClientFinalizeResult], which consists of an
-//! output for the protocol.
+//! [VerifiableClient::finalize] to produce an output for the protocol.
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! # use voprf::VerifiableClient;
 //! # use rand::{rngs::OsRng, RngCore};
 //! #
 //! # let mut client_rng = OsRng;
-//! # let client_blind_result = VerifiableClient::<Default>::blind(
-//! #     b"input",
+//! # let client_blind_result = VerifiableClient::<Group, Hash>::blind(
+//! #     b"input".to_vec(),
 //! #     &mut client_rng,
 //! # ).expect("Unable to construct client");
 //! # use voprf::VerifiableServer;
 //! # let mut server_rng = OsRng;
-//! # let server = VerifiableServer::<Default>::new(&mut server_rng)
+//! # let server = VerifiableServer::<Group, Hash>::new(&mut server_rng)
 //! #   .expect("Unable to construct server");
 //! # let server_evaluate_result = server.evaluate(
 //! #     &mut server_rng,
 //! #     client_blind_result.message,
-//! #     &Metadata::none(),
+//! #     None,
 //! # ).expect("Unable to perform server evaluate");
-//! use voprf::Metadata;
 //! let client_finalize_result = client_blind_result.state.finalize(
 //!     server_evaluate_result.message,
 //!     server_evaluate_result.proof,
 //!     server.get_public_key(),
-//!     &Metadata::none(),
+//!     None,
 //! ).expect("Unable to perform client finalization");
 //!
-//! println!("VOPRF output: {:?}", client_finalize_result.output.to_vec());
+//! println!("VOPRF output: {:?}", client_finalize_result.to_vec());
 //! ```
 //!
 //! # Advanced Usage
@@ -333,12 +289,8 @@
 //! states and messages:
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! # use voprf::VerifiableClient;
 //! # use rand::{rngs::OsRng, RngCore};
 //! #
@@ -346,8 +298,8 @@
 //! let mut client_states = vec![];
 //! let mut client_messages = vec![];
 //! for _ in 0..10 {
-//!     let client_blind_result = VerifiableClient::<Default>::blind(
-//!         b"input",
+//!     let client_blind_result = VerifiableClient::<Group, Hash>::blind(
+//!         b"input".to_vec(),
 //!         &mut client_rng,
 //!     ).expect("Unable to construct client");
 //!     client_states.push(client_blind_result.state);
@@ -361,12 +313,8 @@
 //! along with a single proof:
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! # use voprf::VerifiableClient;
 //! # use rand::{rngs::OsRng, RngCore};
 //! #
@@ -374,38 +322,32 @@
 //! # let mut client_states = vec![];
 //! # let mut client_messages = vec![];
 //! # for _ in 0..10 {
-//! #     let client_blind_result = VerifiableClient::<Default>::blind(
-//! #         b"input",
+//! #     let client_blind_result = VerifiableClient::<Group, Hash>::blind(
+//! #         b"input".to_vec(),
 //! #        &mut client_rng,
 //! #     ).expect("Unable to construct client");
 //! #     client_states.push(client_blind_result.state);
 //! #     client_messages.push(client_blind_result.message);
 //! # }
-//! # use voprf::Metadata;
 //! # use voprf::VerifiableServer;
 //! let mut server_rng = OsRng;
-//! # let server = VerifiableServer::<Default>::new(&mut server_rng)
+//! # let server = VerifiableServer::<Group, Hash>::new(&mut server_rng)
 //! #   .expect("Unable to construct server");
 //! let server_batch_evaluate_result = server.batch_evaluate(
 //!     &mut server_rng,
 //!     &client_messages,
-//!     &Metadata::none(),
+//!     None,
 //! ).expect("Unable to perform server batch evaluate");
 //! ```
 //!
 //! Then, the client calls [VerifiableClient::batch_finalize] on
 //! the client states saved from the first step, along with the messages
-//! returned by the server (constructing a [BatchFinalizeInput]), along with the
-//! server's proof, in order to produce a vector of outputs if the proof
-//! verifies correctly.
+//! returned by the server, along with the server's proof, in order to produce
+//! a vector of outputs if the proof verifies correctly.
 //!
 //! ```
-//! # use voprf::CipherSuite;
-//! # struct Default;
-//! # impl CipherSuite for Default {
-//! #     type Group = curve25519_dalek::ristretto::RistrettoPoint;
-//! #     type Hash = sha2::Sha512;
-//! # }
+//! # type Group = curve25519_dalek::ristretto::RistrettoPoint;
+//! # type Hash = sha2::Sha512;
 //! # use voprf::VerifiableClient;
 //! # use rand::{rngs::OsRng, RngCore};
 //! #
@@ -413,36 +355,31 @@
 //! # let mut client_states = vec![];
 //! # let mut client_messages = vec![];
 //! # for _ in 0..10 {
-//! #     let client_blind_result = VerifiableClient::<Default>::blind(
-//! #         b"input",
+//! #     let client_blind_result = VerifiableClient::<Group, Hash>::blind(
+//! #         b"input".to_vec(),
 //! #        &mut client_rng,
 //! #     ).expect("Unable to construct client");
 //! #     client_states.push(client_blind_result.state);
 //! #     client_messages.push(client_blind_result.message);
 //! # }
-//! # use voprf::Metadata;
 //! # use voprf::VerifiableServer;
-//! use voprf::BatchFinalizeInput;
 //! let mut server_rng = OsRng;
-//! # let server = VerifiableServer::<Default>::new(&mut server_rng)
+//! # let server = VerifiableServer::<Group, Hash>::new(&mut server_rng)
 //! #   .expect("Unable to construct server");
 //! # let server_batch_evaluate_result = server.batch_evaluate(
 //! #     &mut server_rng,
 //! #     &client_messages,
-//! #     &Metadata::none(),
+//! #     None,
 //! # ).expect("Unable to perform server batch evaluate");
-//! let batch_finalize_input = BatchFinalizeInput::new(
-//!     client_states,
-//!     server_batch_evaluate_result.messages,
-//! );
 //! let client_batch_finalize_result = VerifiableClient::batch_finalize(
-//!     batch_finalize_input,
+//!     &client_states,
+//!     &server_batch_evaluate_result.messages,
 //!     server_batch_evaluate_result.proof,
 //!     server.get_public_key(),
-//!     &Metadata::none(),
+//!     None,
 //! ).expect("Unable to perform client batch finalization");
 //!
-//! println!("VOPRF batch outputs: {:?}", client_batch_finalize_result.outputs);
+//! println!("VOPRF batch outputs: {:?}", client_batch_finalize_result);
 //! ```
 //!
 //! ## Metadata
@@ -454,35 +391,45 @@
 //! This metadata can be constructed with some type of higher-level domain separation
 //! to avoid cross-protocol attacks or related issues.
 //!
-//! The default metadata simply consists of the empty vector of bytes, but a custom
-//! metadata can be specified, for example, by: `Metadata(b"custom metadata")`.
+//! A custom metadata can be specified, for example, by: `Some(b"custom metadata")`.
 //!
 //! # Features
 //!
-//! - The `p256` feature enables using p256 as the underlying group for the [CipherSuite] choice.
+//! - The `p256` feature enables using p256 as the underlying group for the [Group](group::Group) choice.
 //! Note that this is currently an experimental feature ⚠️, and is not yet ready for production use.
 //!
-//! - The `serialize` feature, enabled by default, provides convenience functions for serializing and deserializing with
+//! - The `serde` feature, enabled by default, provides convenience functions for serializing and deserializing with
 //! [serde](https://serde.rs/).
 //!
-//! - The `u32_backend` and `u64_backend` features are re-exported from
+//! - The `danger` feature, disabled by default, exposes functions for setting and getting
+//!   internal values not available in the default API. These functions are intended for use in
+//!   by higher-level cryptographic protocols that need access to these raw values and are able to
+//!   perform the necessary validations on them (such as being valid group elements).
+//!
+//! - The backend features are re-exported from
 //! [curve25519-dalek](https://doc.dalek.rs/curve25519_dalek/index.html#backends-and-features) and allow for selecting
-//! the corresponding backend for the curve arithmetic used. The `u64_backend` feature is included as the default.
+//! the corresponding backend for the curve arithmetic used. The `ristretto255_u64` feature is included as the default.
+//! Other features are mapped as `ristretto255_u32`, `ristretto255_fiat_u64` and `ristretto255_fiat_u32`.
+//!
+//! - The `ristretto255_simd` feature is re-exported from
+//! [curve25519-dalek](https://doc.dalek.rs/curve25519_dalek/index.html#backends-and-features) and enables parallel formulas,
+//! using either AVX2 or AVX512-IFMA. This will automatically enable the `ristretto255_u64` feature and requires Rust nightly.
 
-#![cfg_attr(not(feature = "bench"), deny(missing_docs))]
 #![deny(unsafe_code)]
+#![warn(clippy::cargo, missing_docs)]
+#![allow(clippy::multiple_crate_versions)]
 #![cfg_attr(not(feature = "std"), no_std)]
+#![cfg_attr(docsrs, feature(doc_cfg))]
 
 extern crate alloc;
 
 #[macro_use]
 mod impls;
 #[macro_use]
-mod serialization;
-mod ciphersuite;
+mod util;
 pub mod errors;
 pub mod group;
-pub mod hash;
+mod serialization;
 mod voprf;
 
 #[cfg(test)]
@@ -490,12 +437,8 @@ mod tests;
 
 // Exports
 
-pub use rand;
-
-pub use crate::ciphersuite::CipherSuite;
 pub use crate::voprf::{
-    BatchFinalizeInput, BlindedElement, EvaluationElement, Metadata, NonVerifiableClient,
-    NonVerifiableClientBlindResult, NonVerifiableClientFinalizeResult, NonVerifiableServer,
-    NonVerifiableServerEvaluateResult, VerifiableClient, VerifiableClientBlindResult,
-    VerifiableClientFinalizeResult, VerifiableServer, VerifiableServerEvaluateResult,
+    BlindedElement, EvaluationElement, NonVerifiableClient, NonVerifiableClientBlindResult,
+    NonVerifiableServer, NonVerifiableServerEvaluateResult, VerifiableClient,
+    VerifiableClientBlindResult, VerifiableServer, VerifiableServerEvaluateResult,
 };
